@@ -236,6 +236,15 @@ let add_set ~kind ~is_exn id ~set =
 (* General purpose code generation module *)
 
 module Gen = struct
+
+  (* same conversion as camlp4 does when converting its ast into ocaml's ast *)
+  let regular_constr_of_revised_constr = function
+    | " True" -> "True"
+    | " False" -> "False"
+    | "True" -> "true"
+    | "False" -> "false"
+    | s -> s
+
   (* Map of record field source locations to their default expression *)
   let record_defaults : (Loc.t, Ast.expr) Hashtbl.t = Hashtbl.create 0
 
@@ -274,7 +283,7 @@ module Gen = struct
       <:patt@loc< $x$ | $paOr_of_list xs$ >>
 
   module PP = Camlp4.Printers.OCaml.Make (Syntax)
-  let conv_ctyp = (new PP.printer ())#ctyp
+  let conv_ctyp = (new PP.printer ~comments:false ())#ctyp
 
   let string_of_ctyp ctyp =
     try
@@ -359,16 +368,18 @@ module Gen = struct
   exception Stop
   let type_is_recursive short_circuit type_name = object (self)
     inherit fold as super
-    method ctyp ctyp =
+    method! ctyp ctyp =
       match short_circuit ctyp with
       | Some false -> self
       | Some true -> raise Stop
       | None ->
         match ctyp with
-        | <:ctyp< $_$ : $ctyp$ >> ->
+        | <:ctyp< $lid:_$ : $ctyp$ >> ->
           (* or else we would say that [type t = { t : int }] is recursive *)
           self#ctyp ctyp
         | <:ctyp< $lid:id$ >> -> if id = type_name then raise Stop else self
+        | <:ctyp< $uid:_$ : $args$ -> $_return_type$ >> -> self#ctyp args
+        | <:ctyp< $uid:_$ : $_return_type$ >> -> self
         | ctyp -> super#ctyp ctyp
   end
   let type_is_recursive ?(stop_on_functions = true) ?(short_circuit = fun _ -> None) type_name tp =
@@ -406,12 +417,12 @@ module Gen = struct
 
   let ignore_everything = object (self)
     inherit map as super
-    method sig_item sig_item =
+    method! sig_item sig_item =
       match super#sig_item sig_item with
       | <:sig_item@loc< value $id$ : $ctyp$ >> ->
         <:sig_item@loc< value $id$ : _no_unused_value_warning_ $ctyp$ >>
       | sig_item -> sig_item
-    method str_item str_item =
+    method! str_item str_item =
       match super#str_item str_item with
       | <:str_item@loc< value $rec:_$ $bindings$ >> as str_item -> (
         match self#ignore_binding bindings with
@@ -566,8 +577,8 @@ end = struct
     val bound_names = []
     method bound_names = bound_names
     method! ctyp = function
-      | Ast.TyDcl (_loc, n, _tpl, _tk, _cl) ->
-        {< bound_names = n :: bound_names >}
+      | Ast.TyDcl (_loc, n, tpl, tk, _cl) ->
+        {< bound_names = (n, tpl, tk) :: bound_names >}
       | ctyp ->
         super#ctyp ctyp
   end
@@ -598,7 +609,7 @@ end = struct
         | Some (id, _loc, params) ->
           let id =
             try
-              let new_ = StringMap.find id bound in
+              let new_, _, _ = StringMap.find id bound in
               used_bound := StringMap.add id (_loc, List.length params) !used_bound;
               new_
             with Not_found -> id in
@@ -615,15 +626,16 @@ end = struct
   let referenced_names td =
     let bound_names = bound_names td in
     let bound_names_map =
-      List.fold_left (fun acc name -> StringMap.add name (gen ()) acc)
+      List.fold_left (fun acc (name, tpl, tk) ->
+        StringMap.add name (gen (), tpl, tk) acc)
         StringMap.empty bound_names in
     let used_bound = ref StringMap.empty in
     let td = (referenced_names used_bound bound_names_map)#ctyp td in
     let bound_names_map =
-      StringMap.fold (fun key v acc ->
+      StringMap.fold (fun key (v, tpl, tk) acc ->
         try
           let arity = StringMap.find key !used_bound in
-          StringMap.add key (v, arity) acc
+          StringMap.add key (v, arity, tpl, tk) acc
         with Not_found -> acc
       ) bound_names_map StringMap.empty in
     td, bound_names_map, used_bound
@@ -640,9 +652,18 @@ end = struct
 
   let build_common _loc td =
     let td2, map, _set = referenced_names td in
-    StringMap.fold (fun k (v, arity) acc ->
+    StringMap.fold (fun k (v, arity, tpl, tk) acc ->
       let tydcl =
-        TyDcl (_loc, v, params_of_arity arity, constructor_of_arity k arity, [])
+        let tpl, rhs =
+          match tk with
+          | <:ctyp< $_$ == $_$ >> ->
+            (* Here we use the fact that when saying type nonrec ('a, 'b) t = ('a, 'b) t = ...,
+               the two list of parameters must be the same (not even shuffling one list is
+               allowed). *)
+            tpl, tk
+          | _ -> params_of_arity arity, constructor_of_arity k arity
+        in
+        TyDcl (_loc, v, tpl, rhs, [])
       in
       let new_constraints =
         <:with_constr< type $constructor_of_arity v arity$ := $constructor_of_arity k arity$ >>
@@ -900,11 +921,11 @@ let () =
      module M : sig type t = [ `A ] with sexp end = ...
    is likely to generate a warning 'unused value t_of_sexp__' in the signature (the same
    warning in an implementation would be already removed).
-   To work around that, for every 'val name : type' auto generated, we insert a
-   'val name : _no_unused_value_warning_' next to it.
+   To work around that, every auto generated 'val name : type' is replaced by
+   'val name : type _no_unused_value_warning_'.
    And in a second step (could probably be done in one step, but it would be complicated),
-   we try to generate an expression that will use these name (which we recognize thanks to
-   the '_no_unused_value_warning_' mark).
+   we try to generate an expression that will use these names (which we recognize because
+   of the mark), and whether we succeed or not, we remove the mark.
    To use a 'val name : type' in a context like:
      module M : sig val name : type end = ...
    you simply need to do:
@@ -928,7 +949,7 @@ let use_idents_in loc idents body =
 let ignore = object (self)
   inherit Ast.map as super
 
-  method expr = function
+  method! expr = function
     | <:expr@loc< let module $uid:m$ : $module_type$ = $module_expr$ in $body$ >> ->
       let module_expr = self#module_expr module_expr in
       let idents, module_type = self#ignore_module_type module_type in
@@ -937,7 +958,7 @@ let ignore = object (self)
       <:expr@loc< let module $uid:m$ : $module_type$ = $module_expr$ in $body$ >>
     | expr -> super#expr expr
 
-  method str_item = function
+  method! str_item = function
     | <:str_item@loc< module type $s$ = $module_type$ >> ->
       let idents, module_type = self#ignore_module_type module_type in
       let warnings_removal = use_idents loc (qualify_idents loc s idents) in
@@ -959,6 +980,12 @@ let ignore = object (self)
       let warnings_removal = use_idents loc (qualify_idents loc m idents) in
       <:str_item@loc< module $uid:m$ : $module_type$ = $module_expr$; $warnings_removal$ >>
 
+    | <:str_item@loc< include ($module_expr$ : $module_type$) >> ->
+      let module_expr = self#module_expr module_expr in
+      let idents, module_type = self#ignore_module_type module_type in
+      let warnings_removal = use_idents loc idents in
+      <:str_item@loc< include ($module_expr$ : $module_type$); $warnings_removal$ >>
+
     | StMod _
     | StSem _
     | StInc _
@@ -997,7 +1024,7 @@ let ignore = object (self)
            $warnings_removal$ *)
         me
 
-  method module_expr = function
+  method! module_expr = function
     | MeFun (loc, _, _, _) as me -> self#fold_map_on_functor_arg <:str_item@loc< >> me
 
     | MeStr _
@@ -1010,7 +1037,7 @@ let ignore = object (self)
       super#module_expr me
 
   (* Strip all the 'markers' that have not been handled *)
-  method sig_item = function
+  method! sig_item = function
     | <:sig_item@loc< value $id$ : _no_unused_value_warning_ $ctyp$  >> ->
       <:sig_item@loc< value $id$ : $ctyp$ >>
     | sig_item -> super#sig_item sig_item
@@ -1061,7 +1088,7 @@ end
 let strip = object
   inherit Ast.map as super
 
-  method sig_item = function
+  method! sig_item = function
   | <:sig_item@loc< value $id$ : _no_unused_value_warning_ $ctyp$  >> ->
     <:sig_item@loc< value $id$ : $ctyp$ >>
   | sig_item -> super#sig_item sig_item
